import datetime
import os
import json
import serial
import time
import traceback

from functools import wraps

from .sys_utils import ShellCommand
from . import settings as S


PROVIDER_ERROR = 0
PROVIDER_OK = 1
PROVIDER_NEW = 2

WB_MODULES_4G = ['wbc-4g', 'wbc-4g-usb']

MODEM_TYPE = 'type'
MODEM_SIMSELECT_GPIO = 'simselect_gpio'
MODEM_STATUS_GPIO = 'status_gpio'
MODEM_SIM_DATA = 'sim_data'
MODEM_SIM_SLOT = 'sim_slot'
MODEM_STATUS = 'status'

STATE_SIM_UNKNOWN = 0
STATE_SIM_NOT_INSERTED = 1
STATE_SIM_DATA = 2

STATE_TRUE = True
STATE_FALSE = False

AT_COMMAND_ERROR_SIM_NOT_INSERTED = "+CME ERROR: SIM not inserted".lower()

COMMAND_MODEM_TEMPLATE = ". /usr/lib/wb-utils/wb_env.sh; " \
                         "wb_source of; of_has_prop \"aliases\" \"wbc_modem\" && " \
                         "modem_node=$(of_get_prop_str \"aliases\" \"wbc_modem\") || modem_node=\"/wirenboard/gsm\"; " \
                         "of_get_prop_gpionum $modem_node {0}"

GPIO_FILEPATH_TEMPLATE = "/sys/class/gpio/gpio{0}/value"


# def simselect_log(function):
#     @wraps(function)
#     def wrapper(*args, **kwargs):
#         log_str = "SET sim slot {0} -> {1}".format(args[0]._sim_slot, args[1])
#         args[0].logger.info(log_str)
#         result = function(args[0], args[1], *args, **kwargs)
#         args[0].logger.info(log_str + " ... OK")
#         return result
#     return wrapper


class SIMData:
    def __init__(self):
        self._data = {}
        self._state = STATE_SIM_UNKNOWN

    def set_unknown(self):
        self._state = STATE_SIM_UNKNOWN

    def set_not_inserted(self):
        self._state = STATE_SIM_NOT_INSERTED

    def set_data(self, sim_data):
        if sim_data is None:
            self.set_unknown()

        error_flag = False
        iccid = sim_data.get('iccid')
        provider = sim_data.get('provider')

        ppp_provider = S.PPP_PROVIDER_DEFAULT
        if provider is not None:
            for p in S.PPP_PROVIDER_LIST:
                if p in provider.lower():
                    ppp_provider = p
                    break

        if (iccid is None) or (provider is None):
            error_flag = True

        self._state = STATE_SIM_DATA
        created_at = datetime.datetime.now()
        self._data = {
            'iccid': iccid,
            'provider': provider,
            'ppp_provider': ppp_provider,
            'error': error_flag,
            'created_at': str(created_at),
            'created_at_ts': created_at.timestamp()
        }

    def is_error(self):
        return self._data.get('error', False)

    def get_data(self):
        result = self._data.copy()
        result['state'] = self._state
        return result

    def get_ppp_provider(self):
        return self._data.get('ppp_provider', S.PPP_PROVIDER_DEFAULT)

    def is_data(self):
        return self._state == STATE_SIM_DATA

    def is_not_inserted(self):
        return self._state == STATE_SIM_NOT_INSERTED

    def is_unknown(self):
        return self._state == STATE_SIM_UNKNOWN


class GSMModem(ShellCommand):
    def __init__(self, logger=None):
        super().__init__(logger)

        self._status = False
        self._modem_type = S.MODEM_TYPE_4G

        self._status_gpio = None
        self._simselect_gpio = None

        self._slots_sim_data = {
            0: SIMData(),
            1: SIMData()
        }

        self._sim_slot = 0

        self._current_ppp_provider = S.PPP_PROVIDER_DEFAULT

    def settings(self):
        slots_sim_data = {k: v.get_data() for k, v in self._slots_sim_data.items()}
        return {
            MODEM_STATUS: self._status,
            MODEM_TYPE: self._modem_type,
            MODEM_STATUS_GPIO: self._status_gpio,
            MODEM_SIMSELECT_GPIO: self._simselect_gpio,
            MODEM_SIM_DATA: slots_sim_data,
            MODEM_SIM_SLOT: self._sim_slot
        }

    def _check_gpio_status(self):
        if self._status_gpio is None:
            self._status_gpio = self._get_gpio('status-gpios')
            self.logger.info("self._status_gpio = \"{0}\"".format(self._status_gpio))
            if self._status_gpio is None:
                return None
        filepath = GPIO_FILEPATH_TEMPLATE.format(self._status_gpio)
        self.logger.info("filepath = \"{0}\"".format(filepath))
        if not os.path.isfile(filepath):
            return None
        cmd = "cat {0}".format(filepath)
        status_str = self.command_strip_output(cmd)
        self.logger.info("GPIO modem status data \"{0}\"".format(status_str))
        return True if status_str == '1' else False

    def _check_gpio_simselect(self):
        if self._simselect_gpio is None:
            self._simselect_gpio = self._get_gpio('simselect-gpios')
            if self._simselect_gpio is None:
                return False
        filepath = GPIO_FILEPATH_TEMPLATE.format(self._simselect_gpio)
        if not os.path.isfile(filepath):
            return False
        cmd = "cat {0}".format(filepath)
        simselect_str = self.command_strip_output(cmd)
        self.logger.info("GPIO modem simselect data \"{0}\"".format(simselect_str))
        return True

    def check_status(self):
        if not self._status:
            # self._restart_hard()
            if not self._restart_hard():
                # self._status = False
                return self._status

        gpio_status = self._check_gpio_status()
        if gpio_status is None:
            return self._status
        elif not gpio_status:
            self._status = False
            return self._status

        # Here gpio_status == True
        if not self._status:
            self._initialize()
        self._status = True
        return self._status

    def set_status(self, status):
        self._status = status

    def skip(self):
        if not self._status:
            return True
        return False

    def _initialize(self):
        time.sleep(S.DELAY_COMMAND)

        self._modem_type = self._get_type()
        self.logger.info("Modem dev \"{0}\" exists - {1}".format(S.MODEM_DEV_DEFAULT_4G,
                                                                 os.path.exists(S.MODEM_DEV_DEFAULT_4G)))
        self._current_ppp_provider = S.PPP_PROVIDER_DEFAULT

        gpio_simselect = self._check_gpio_simselect()
        if not gpio_simselect:
            return

        _slots_sim_data = self._get_slots_sim_data()
        if _slots_sim_data is not None:
            self._slots_sim_data = _slots_sim_data

        sim_data = self._slots_sim_data.get(self._sim_slot)
        if sim_data is not None:
            self._current_ppp_provider = sim_data.get_ppp_provider()
        else:
            next_sim_slot = self._next_sim_slot()
            sim_data = self._slots_sim_data.get(next_sim_slot)
            if sim_data is not None:
                self._current_ppp_provider = sim_data.get_ppp_provider()

    def _get_slots_sim_data(self):
        ser = self._open_serial_port()
        if ser is None:
            return None

        result = {
            0: SIMData(),
            1: SIMData()
        }
        for _sim_slot in [1, 0]:
            set_result, sim_not_inserted = self._set_sim_slot_serial(_sim_slot, ser)
            if not set_result:
                continue
            if sim_not_inserted:
                result[_sim_slot].set_not_inserted()
            else:
                result[_sim_slot] = self._sim_data_serial(ser)
        ser.close()
        return result

    def _next_sim_slot(self):
        return 1 if self._sim_slot == 0 else 0

    # def _get_sim_provider(self, sim_data):
    #     if sim_data is None:
    #         return self._current_ppp_provider
    #     return sim_data.get("ppp_provider", S.PPP_PROVIDER_DEFAULT)

    def sim_not_inserted(self):
        sim_data = self._slots_sim_data.get(self._sim_slot)
        if (sim_data is not None) and (sim_data.is_not_inserted()):
            return True
        return False

    def current_ppp_provider(self):
        return self._current_ppp_provider

###
    # def _current_sim_data(self):
    #     if self._slots_sim_data is None:
    #         return None
    #     return self._slots_sim_data.get(self._sim_slot)
    #
    # def _update_current_sim_data(self, sim_data):
    #     if self._slots_sim_data is None:
    #         self._slots_sim_data = {}
    #     if sim_data is None:
    #         self._slots_sim_data.pop(self._sim_slot, None)
    #     else:
    #         self._slots_sim_data[self._sim_slot] = sim_data
###

    def _get_gpio(self, gpio_name):
        cmd = COMMAND_MODEM_TEMPLATE.format(gpio_name)
        gpio = self.command_strip_output(cmd)
        if not gpio:
            self.logger.error("GPIO \"{0}\" is empty".format(gpio_name))
            gpio = None
        return gpio

    def change_sim_slot(self):
        gpio_simselect = self._check_gpio_simselect()
        if not gpio_simselect:
            return False

        new_sim_slot = self._next_sim_slot()

        # result = False
        # ser = self._open_serial_port()
        # if ser is not None:
        #     result, sim_not_inserted = self._set_sim_slot_serial(new_sim_slot, ser)
        #     ser.close()
        # if not result:
        #     result = self._set_sim_slot_hard(new_sim_slot)
        result = self._set_sim_slot_hard(new_sim_slot)
        return result

    def update_sim_slot_data(self):
        is_ppp_provider_changed = False
        self.logger.info("CHECK \"{0}\" sim slot data".format(self._sim_slot))
        current_sim_data = self._slots_sim_data.get(self._sim_slot)
        if (current_sim_data is not None) and current_sim_data.is_data() and current_sim_data.is_error():
            self.logger.warning(
                "CHECK \"{0}\" sim slot data:\n - bad CURRENT sim data = {1}".format(self._sim_slot,
                                                                                     current_sim_data.get_data()))

        ser = self._open_serial_port()
        if ser is None:
            return is_ppp_provider_changed
        sim_data = self._sim_data_serial(ser)
        ser.close()

        if sim_data.is_unknown():
            return is_ppp_provider_changed

        if sim_data.is_not_inserted():  # SIM not inserted
            self._slots_sim_data[self._sim_slot] = sim_data
            return is_ppp_provider_changed

        if sim_data.is_data():
            if sim_data.is_error():
                self.logger.warning(
                    "CHECK \"{0}\" sim slot data:\n - bad NEW sim data = {1}".format(self._sim_slot,
                                                                                     sim_data.get_data()))
                return is_ppp_provider_changed
            else:
                new_ppp_provider = sim_data.get_ppp_provider()
                is_ppp_provider_changed = self._current_ppp_provider != new_ppp_provider
                if is_ppp_provider_changed:
                    self.logger.warning("CHECK \"{0}\" sim slot data - UPDATE:\n - old = {1}\n - new = {2}"
                                        .format(self._sim_slot, self._current_ppp_provider, new_ppp_provider))
                    self._current_ppp_provider = new_ppp_provider
                self._slots_sim_data[self._sim_slot] = sim_data
        return is_ppp_provider_changed

    def _set_simselect_gpio_value(self, sim_slot):
        filepath = GPIO_FILEPATH_TEMPLATE.format(self._simselect_gpio)
        if not os.path.exists(filepath):
            return False
        self.command_exec("echo {0} > {1}".format(sim_slot, filepath))
        time.sleep(S.DELAY_COMMAND)
        return True

    def _set_sim_slot_hard(self, sim_slot):
        log_str = "HARD SET sim slot {0} -> {1}".format(self._sim_slot, sim_slot)
        self.logger.info(log_str)

        if not self._set_simselect_gpio_value(sim_slot):
            self.logger.info(log_str + " ... ERROR")
            return False
        self.off()
        self.on()
        self._sim_slot = sim_slot
        self.logger.info(log_str + " ... OK")
        return True

    # @simselect_log
    def _set_sim_slot_serial(self, sim_slot, ser):
        log_str = "SET sim slot {0} -> {1}".format(self._sim_slot, sim_slot)
        self.logger.info(log_str)

        sim_not_inserted = False
        if not self._set_simselect_gpio_value(sim_slot):
            self.logger.info(log_str + " ... ERROR")
            return False, sim_not_inserted

        cmd = "AT+CFUN=0"
        msg = self.at_command_exec(cmd, ser)

        # time.sleep(S.DELAY_COMMAND_RESULT)

        cmd = "AT+CFUN=1"
        msg = self.at_command_exec(cmd, ser)
        if msg is not None:
            r = self._parse_CFUN1(msg)
            if r is None:  # SIM not inserted
                sim_not_inserted = True

        self._sim_slot = sim_slot
        time.sleep(S.DELAY_GSM)
        self.logger.info(log_str + " ... OK")
        return True, sim_not_inserted

    def _sim_data_serial(self, ser):
        result = SIMData()
        try:
            cmd = "AT+CCID"
            msg = self.at_command_exec(cmd, ser)
            if msg is None:
                iccid = None
            else:
                iccid = self._parse_CCID(msg)
                if iccid is None:  # SIM not inserted
                    result.set_not_inserted()
                    return result
        except:
            self.logger.error(traceback.format_exc())
            iccid = None

        try:
            cmd = "AT+CSPN?"
            msg = self.at_command_exec(cmd, ser)
            provider_name = self._parse_CSPN(msg)
        except:
            self.logger.error(traceback.format_exc())
            provider_name = None

        # TODO AT COMMANDS:
        # "AT+CIMI"
        # "AT+COPS=3,0" + "AT+COPS?" - operator_name

        result.set_data({
            'iccid': iccid,
            'provider': provider_name
        })
        return result

    @staticmethod
    def _msg_lines(msg):
        # if msg is None:
        #     return None
        return [p.strip() for p in msg.splitlines()]

    def _parse_CSPN(self, msg):
        try:
            result_start = '+CSPN:'
            parts = self._msg_lines(msg)
            for p in parts:
                if p.startswith(result_start):
                    return p[len(result_start):].split('"')[1]
                    # return p.split()[1].split(',')[0][1:-1]

            # if p[-1].lower() == 'ok':
            #     provider_name = p[-3].split('"')[-2]
            #     return provider_name
        except:
            self.logger.info(traceback.format_exc())
        return msg

    @staticmethod
    def _check_at_sim_not_inserted(parts):
        for p in parts:
            if AT_COMMAND_ERROR_SIM_NOT_INSERTED in p.lower():
                return True
        return False

    def _parse_CCID(self, msg):
        try:
            parts = self._msg_lines(msg)
            if self._check_at_sim_not_inserted(parts):
                return None

            result_start = '+CCID:'
            for p in parts:
                if p.startswith(result_start):
                    return p[len(result_start):].strip()
                    # return p.split()[1]

            # if p[-1].lower() == 'ok':
            #     provider_name = p[-3].split()[-1]
            #     return provider_name
        except:
            self.logger.info(traceback.format_exc())
        return msg

    def _parse_CFUN1(self, msg):
        try:
            parts = self._msg_lines(msg)
            if self._check_at_sim_not_inserted(parts):
                return None
        except:
            self.logger.info(traceback.format_exc())
        return msg

    def _parse_CSQ(self, msg):
        try:
            parts = msg.splitlines()
            for p in parts:
                if p.startswith('+CSQ:'):
                    return p.split()[1]
            # if p[-1].lower() == 'ok':
            #     provider_name = p[-3].split()[-1]
            #     return provider_name
        except:
            self.logger.warning("Can't parse AT+CSQ output \"{0}\":\n{1}".format(msg, traceback.format_exc()))
        return None

    def signal_quality(self):
        # if self._settings[MODEM_DEV_FILEPATH] == S.MODEM_DEV_DEFAULT:
        #     return None
        result = {'sim slot': self._sim_slot}
        signal_quality = self._signal_quality()
        if (signal_quality is not None) and signal_quality:
            result['signal quality'] = signal_quality
        return result

    def _signal_quality(self):
        cmd = 'AT+CSQ'
        msg = self.at_command_exec(cmd)
        csq = self._parse_CSQ(msg)
        return csq

    def at_command_exec(self, cmd, ser=None):
        if ser is None:
            _ser = self._open_serial_port()
            if _ser is None:
                return None
        else:
            _ser = ser
        try:
            _cmd = "{0}\r\n".format(cmd.strip())
            time.sleep(S.DELAY_COMMAND)
            _ser.read(_ser.in_waiting)
            _ser.write(_cmd.encode())
            time.sleep(S.DELAY_COMMAND_RESULT)
            msg = _ser.read(_ser.in_waiting).decode("utf-8")
        except Exception:
            self.logger.error(traceback.format_exc())
            msg = None
        if ser is None:
            _ser.close()

        if msg is not None:
            self.logger.info("AT COMMAND \"{0}\":\n\"{1}\"".format(cmd, msg))
        else:
            self.logger.error("AT COMMAND \"{0}\": \"{1}\"".format(cmd, msg))
        return msg

    def _open_serial_port(self):
        try:
            ser = serial.Serial(port=S.MODEM_DEV_AT_COMMAND,  # S.MODEM_DEV_DEFAULT,
                                baudrate=115200, bytesize=8, parity='N', stopbits=1, timeout=1)
        except:
            self.logger.error("OPEN SERIAL PORT:\n{0}".format(traceback.format_exc()))
            return None
        try:
            cmd = "AAAAAAAAAT"
            msg = self.at_command_exec(cmd, ser)
            # time.sleep(S.DELAY_COMMAND)
        except Exception:
            self.logger.error(traceback.format_exc())
        return ser

#################################################

    # $ wb-gsm on - включение устройства
    def on(self):
        self.command_exec("/usr/bin/wb-gsm on")
        time.sleep(S.DELAY_GSM)

    def _restart_hard(self):
        self.logger.info("HARD RESTART")
        start_dt = datetime.datetime.now()
        name_pattern = "wb-gsm restart_if_broken"
        self.logger.info("Start \"{0}\"".format(name_pattern))
        self.command_exec_background("/usr/bin/{0}".format(name_pattern))
        self.logger.info("Start \"{0}\" ... OK".format(name_pattern))
        time.sleep(S.DELAY_GSM)

        while (datetime.datetime.now() - start_dt).total_seconds() < (60 + S.DELAY_GSM):
            n = self._wb_process_count(name_pattern)
            if (n is None) or (n > 0):
                self.logger.warning("WAIT \"{0}\", count = {1}".format(name_pattern, n))
                time.sleep(S.DELAY_GSM)
            elif n == 0:
                self.logger.info("PROCESS \"{0}\" finished, count = {1}".format(name_pattern, n))
                self.logger.info("HARD RESTART ... OK")
                time.sleep(S.DELAY_GSM)
                return True

        self.logger.error("(!) KILL \"{0}\"".format(name_pattern))
        self.command_exec("pkill -f '{0}'".format(name_pattern))
        time.sleep(S.DELAY_COMMAND_RESULT)
        n = self._wb_process_count(name_pattern)
        self.logger.warning("(!) REMAINDER \"{0}\", count = {1}".format(name_pattern, n))
        self.logger.info("HARD RESTART ... ERROR [KILL]")
        return False

    def _wb_process_count(self, name_pattern):
        try:
            n_str = self.command_strip_output("pgrep -c -f '{0}'".format(name_pattern))
            n = int(n_str)
            return n
        except:
            self.logger.error(traceback.format_exc())
        return None

    # $ wb-gsm off - выключение устройства
    def off(self):
        self.command_exec("/usr/bin/wb-gsm off")
        # time.sleep(S.DELAY_GSM)

    # $ wb-gsm restart_if_broken - включение или перезагрузка, если модем не отвечает
    def restart(self):
        ret, stdout, stderr = self.command_exec("/usr/bin/wb-gsm restart_if_broken", timeout=60)
        time.sleep(S.DELAY_GSM)
        if (ret is None) or (ret != 0):
            return False
        return True

#################################################

    def _get_type(self):
        return S.MODEM_TYPE_4G
        # # if not os.path.isfile(WB_GSM_TYPE_FILEPATH):
        # #     return S.MODEM_TYPE_DEFAULT
        # if self._check_4g_wb_modem_slot_module():
        #     return S.MODEM_TYPE_4G
        #
        # result = S.MODEM_TYPE_DEFAULT
        # try:
        #     with open(WB_GSM_TYPE_FILEPATH) as f:
        #         type_txt = f.read()
        #         if '4g' in type_txt:
        #             result = S.MODEM_TYPE_4G
        #         elif '3g' in type_txt:
        #             result = S.MODEM_TYPE_3G
        # except:
        #     self.logger.error("Can't read WB file [{0}]:\n{1}".format(WB_GSM_TYPE_FILEPATH, traceback.format_exc()))
        # return result

    def _check_4g_wb_modem_slot_module(self):
        try:
            with open(S.WB_HARDWARE_CONF_FILE_PATH) as f:
                json_conf = json.load(f)
                slots = json_conf.get('slots')
                if (slots is None) or (not slots):
                    return False
                for slot in slots:
                    module = slot.get('module')
                    if module is None:
                        continue
                    if module in WB_MODULES_4G:
                        return True
        except:
            self.logger.error("Read \"{0}\" json-file [{0}]:\n{1}".format(S.WB_HARDWARE_CONF_FILE_PATH,
                                                                          traceback.format_exc()))
        return False

#################################################

    # def _get_dev(self, modem_type):
    #     return S.MODEM_DEV_DEFAULT_4G
    #     if modem_type == S.MODEM_TYPE_4G:
    #         return S.MODEM_DEV_DEFAULT_4G
    #     elif modem_type == S.MODEM_TYPE_3G:
    #         return self._modem_dev_3g()  # S.MODEM_DEV_DEFAULT_3G
    #     return S.MODEM_DEV_DEFAULT
    #
    # def _modem_dev_3g(self):
    #     for i in range(8):
    #         modem_dev = S.MODEM_DEV_TEMPLATE_3G.format(i)
    #         if self._is_available_modem_dev_3g(modem_dev):
    #             return modem_dev
    #     return S.MODEM_DEV_DEFAULT_3G
    #
    # def _is_available_modem_dev_3g(self, dev_filepath):
    #     result = False
    #     if os.path.exists(dev_filepath):
    #         cmd = "udevadm info {0} --query=property".format(dev_filepath)
    #         res, out, err = self.command_exec(cmd, log=False)
    #         if res != 0:
    #             return False
    #         try:
    #             id_model = ''
    #             id_vendor = ''
    #             for line in [s.strip() for s in out.splitlines()]:
    #                 # print("line [{0}]".format(line))
    #                 if not line:
    #                     continue
    #                 tmp = line.split('=')
    #                 if len(tmp) == 2:
    #                     key = tmp[0].strip()
    #                     value = tmp[1].strip()
    #                     if key == 'ID_MODEL':
    #                         id_model = value
    #                     elif key == 'ID_VENDOR':
    #                         id_vendor = value
    #                 # print("id_model=[{0}]; id_vendor=[{1}]".format(id_model, id_vendor))
    #             if (id_model == "SIMCOM_PRODUCT") and (id_vendor == "SIMCOM_VENDOR"):
    #                 result = True
    #         except:
    #             result = False
    #     return result
    #
    # def _check_modem_provider(self, provider_name, modem_dev):
    #     ppp_provider_filepath = S.PPP_PROVIDER_FILE_PATH_TEMPLATE.format(provider_name)
    #     dev_filepath_template = "/dev/tty"
    #     new_content = []
    #     replace_flag = True
    #     try:
    #         with open(ppp_provider_filepath) as f:
    #             for line in f:
    #                 line_tmp = line.strip()
    #                 if replace_flag:
    #                     if (not line_tmp) or line_tmp.startswith("#"):
    #                         # To skip IQnergy comments
    #                         continue
    #
    #                     if line_tmp.startswith(dev_filepath_template):
    #                         if line_tmp == modem_dev:
    #                             return PROVIDER_OK
    #                         else:
    #                             line = "# [IQnergy, {0}]\n#{1}\n".format(datetime.now(), line)
    #                             line += "# New modem dev:\n{0}\n".format(modem_dev)
    #                         replace_flag = False
    #                 new_content.append(line)
    #     except:
    #         self.logger.error(traceback.format_exc())
    #         return PROVIDER_ERROR
    #     with open(ppp_provider_filepath, 'w') as file:
    #         file.writelines(new_content)
    #     return PROVIDER_NEW
