#!/usr/bin/python3

"""
Модуль, реализующий переключение сетевых интерфейсов на контроллере

"""

import os
import json
import traceback
import datetime
import threading
import time

from .sys_utils import IQLog, except_log, get_logger
from .mqtt_client import MQTTClient, interface_settings_are_changed
from .net_utils import NetUtils

from . import settings as S


STORAGE_INTERFACE_SETTINGS = 'interface_settings'
STORAGE_INTERFACE_DATA = 'interface_data'
STORAGE_UPDATED_AT = 'updated_at'
# STORAGE_MODEM = 'modem'
STORAGE_PREV = 'previous_data'


class IQNetworkManager(IQLog):
    def __init__(self, name, config):
        self.name = name
        logger_path = os.path.join(config.logging_dir, "{0}.log".format(self.name))
        self.logger = get_logger(logger_path, self.name)
        self.working_dir = config.working_dir
        super().__init__(logger=self.logger)

        self.logger.info("logger path: \"{0}\"".format(logger_path))
        self.logger.info("working path: \"{0}\"".format(self.working_dir))

        storage_path = os.path.join(self.working_dir, "storage.json")
        self._storage = IQStorage(storage_path, self.logger)

        self._background_threads = []

        self._mqtt_client = MQTTClient(self.logger)
        self._net = NetUtils(self._storage, self.logger)

        self._no_internet_counter = 0
        self._start_datetime = datetime.datetime.now()

    @except_log
    def start(self):
        self.logger.info("STORAGE init state:\n{0}".format(self._storage.txt()))

        self._background_thread(self._mqtt_client.start).start()
        self._background_thread(self._engine).start()

        self.logger.info("    > join background threads")
        for background_thread in self._background_threads:
            if background_thread.is_alive():
                background_thread.join()
        self.logger.info("    > join background threads ... OK")

    @except_log
    def _engine(self):
        self._net.check_dns()
        while True:
            try:
                self.logger.info("\n>>> ENGINE TASK")
                self._engine_iter()
                self.logger.info("\n<<< ENGINE TASK\n\n\n")
            except:
                self.logger.critical(traceback.format_exc())
            time.sleep(S.DELAY_DEFAULT)

    def _engine_iter(self):
        checked_eth_interfaces = self._net.internet_via_eth()  # For controller service needs
        if checked_eth_interfaces:
            self.logger.info("(!) ETHERNET CONNECTION at {0}, skip network settings checking"
                             .format(checked_eth_interfaces))
            self._no_internet_counter = 0
            self._result(None, checked_eth_interfaces)
            return

        prev_interfaces = self._interfaces_from_storage()

        # Update Storage
        if self._storage.is_none():
            self.logger.warning("STORAGE IS NONE - INITIALIZE BY DEFAULT SETTINGS")
            self._storage.init_by_default()
            self._check_for_updates()
            interface_settings_flag = True
            interface_data_flag = True
            self.logger.warning("NEED TO UPDATE LINUX INTERFACES by None STORAGE")
        else:
            interface_settings_flag, interface_data_flag = self._check_for_updates()
            if interface_settings_flag and interface_data_flag:
                self.logger.warning("NEED TO UPDATE LINUX INTERFACES by NEW SETTINGS")

        target_interfaces = self._interfaces_from_storage()
        self._net.ifdown_not_actual_interfaces(prev_interfaces, target_interfaces)

        if not self._net.check_modem(target_interfaces):
            self.logger.error("MODEM STATUS = FALSE")
            # dt = (datetime.datetime.now() - self._start_datetime).total_seconds()
            # dt_max = 0  # 2 * 60
            # self.logger.error("MODEM STATUS = FALSE: {0}/{1} (sec)".format(int(dt), dt_max))
            # if dt >= dt_max:
            #     self.logger.warning("HARD (!) MODEM ON")
            #     if not self._net.modem_on():
            #         time.sleep(S.DELAY_DEFAULT)
            #         self._net.modem_off()
            #         time.sleep(S.DELAY_COMMAND_RESULT)
            #         self.logger.warning("HARD (!) MODEM ON ... ERROR [KILL]")
            #     else:
            #         self.logger.warning("HARD (!) MODEM ON ... OK")
            # else:
            #     time.sleep(S.DELAY_DEFAULT)
        else:
            self.logger.error("MODEM STATUS = TRUE")
            if self._net.update_sim_data():
                interface_settings_flag = True
                interface_data_flag = True
                self.logger.warning("NEED TO UPDATE LINUX INTERFACES by NEW PPP PROVIDER")

        self._net.print_routes()

        if interface_settings_flag:
            if interface_data_flag:
                self.logger.warning("(!!!) UPDATE LINUX INTERFACES")
                self._net.update_linux_interfaces()
            self.logger.info("(!) DELETE ALL DEFAULT GATEWAYS")
            self._net.delete_default_gateways()
            self.logger.warning("(!!!) FINISH [UPDATE INTERFACES BY NEW SETTINGS]")
            self._net.print_routes()

        # if not self._storage[STORAGE_INTERFACES]:  .interface_data()
        #     self.logger.info("There is no any actual INTERFACES... Skip this settings")
        #     return

        current_gateways = self._net.get_current_gateways()
        self.logger.info("current_gateways = {0}".format(current_gateways))

        checked_interfaces = []
        for i, m in target_interfaces.items():
            self.logger.info("CHECK \"{0}\": metric = {1}".format(i, m))
            current_metrics = current_gateways.get(i, [])
            check_result = self._net.check_interface(i, current_metrics, m)
            if check_result:
                checked_interfaces.append(i)

        self._net.print_routes()
        self._result(target_interfaces, checked_interfaces)

        if not checked_interfaces:
            self._no_internet_counter += 1
            self.logger.warning("NO INTERNET {0}/{1}".format(self._no_internet_counter, S.NO_INTERNET_MAX_COUNT))

            if self._no_internet_counter > S.NO_INTERNET_MAX_COUNT:
                self._net.controller_reboot()

            # self.logger.info("Sleep {0} sec".format(S.DELAY_NO_INTERNET))
            # time.sleep(S.DELAY_NO_INTERNET)
            # self.logger.info("Sleep {0} sec".format(S.DELAY_DEFAULT))
            # time.sleep(S.DELAY_DEFAULT)
        else:
            self._no_internet_counter = 0

            bad_interface_counter = 0
            for interface in target_interfaces:
                if interface in [S.ETH0_INTERFACE_NAME, S.ETH1_INTERFACE_NAME]:
                    continue
                if interface in checked_interfaces:
                    self.logger.info(" - \"{0}\" interface ... OK".format(interface))
                else:
                    self.logger.info(" - \"{0}\" interface ... NO INTERNET".format(interface))
                    bad_interface_counter += 1

            if bad_interface_counter == 0:
                self.logger.info("Sleep {0} sec".format(S.DELAY_WITH_INTERNET))
                time.sleep(S.DELAY_WITH_INTERNET)

    def _interfaces_from_storage(self):
        interface_data = self._storage.interface_data()
        if interface_data is None:
            return {}
        return {i: m for (i, d, m) in interface_data if i != S.LO_INTERFACE_NAME}

    def _result(self, target_interfaces=None, checked_interfaces=None):
        current_datetime = datetime.datetime.now()
        actual_interfaces = self._net.get_current_gateways()
        result = {
            'event_timestamp': int(current_datetime.timestamp()),
            'datetime': str(current_datetime),  # TEMP FIELD
            'actual_interfaces': actual_interfaces,
            'modem_data': self._net.modem_settings()
        }

        if target_interfaces is not None:
            result['target_interfaces'] = target_interfaces

        if checked_interfaces is not None:
            result.update({
                'current_interfaces': {checked_interfaces[i]: i for i in range(len(checked_interfaces))},  # TEMP FIELD
                'checked_interfaces': checked_interfaces,
            })
            interface_quality = self._net.interface_quality(checked_interfaces)
            if (interface_quality is not None) and interface_quality:
                result['interface_quality'] = interface_quality

        self.logger.info("Publish result:\n{0}".format(result))
        r = self._mqtt_client.publish_result(result)
        self.logger.info("Publish result: {0}".format('OK' if r else 'ERROR'))

    def _check_for_updates(self):
        if not self._mqtt_client.new_settings():
            return False, False

        client_settings = self._mqtt_client.get_settings()
        if client_settings is None:
            return False, False

        interface_settings_flag, interface_data_flag = self._storage.update_interface_settings(client_settings)
        return interface_settings_flag, interface_data_flag

    @except_log
    def _background_thread(self, target, args=(), daemon=True):
        thread_name = target.__name__
        task_thread = threading.Thread(target=target, args=args, name=thread_name)
        task_thread.daemon = daemon
        self._background_threads.append(task_thread)
        self.logger.info("Add \"{0}\" to background threads (length = {1})".format(thread_name,
                                                                                   len(self._background_threads)))
        return task_thread


class IQStorage(IQLog):
    def __init__(self, storage_file_path, logger=None):
        super().__init__(logger)
        self._storage_file_path = storage_file_path

        file_storage = self._get_storage_from_file()
        if file_storage is None:
            self._storage = None  # self._storage_init(S.DEFAULT_INTERFACE_SETTINGS)
            # self._storage = DEFAULT_STORAGE_DATA
            # self._set_storage_file()
        else:
            self._storage = file_storage
            self.logger.info("Set STORAGE from json-file [{0}]".format(self._storage_file_path))

    def is_none(self):
        if self._storage is None:
            return True
        return False

    def init_by_default(self):
        self._storage = self._storage_init(S.DEFAULT_INTERFACE_SETTINGS)
        self._set_storage_file()

    def _storage_init(self, interface_settings):
        try:
            result_storage = {
                STORAGE_INTERFACE_SETTINGS: interface_settings,
                STORAGE_INTERFACE_DATA: self._get_interface_data(interface_settings),
                STORAGE_UPDATED_AT: str(datetime.datetime.now())
            }
        except:
            self.logger.error(traceback.format_exc())
            result_storage = self._storage
        return result_storage

    @except_log
    def _get_storage_from_file(self):
        if os.path.isfile(self._storage_file_path):
            try:
                with open(self._storage_file_path) as f:
                    json_storage = json.load(f)
                return json_storage
            except:
                self.logger.error("Can't read from the STORAGE json-file [{0}]:\n{1}".format(self._storage_file_path,
                                                                                             traceback.format_exc()))
        else:
            self.logger.warning("The STORAGE json-file [{0}] is not file".format(self._storage_file_path))
        return None

    @except_log
    def _set_storage_file(self):
        if ((self._storage is None) or (not isinstance(self._storage, dict))) \
                and os.path.isfile(self._storage_file_path):
            os.remove(self._storage_file_path)
        else:
            try:
                with open(self._storage_file_path, 'w') as f:
                    json.dump(self._storage, f)
            except:
                self.logger.error("Can't write STORAGE json-file [{0}]:\n{1}".format(self._storage_file_path,
                                                                                     traceback.format_exc()))

    @except_log
    def txt(self):
        return json.dumps(self._storage, indent=4, sort_keys=True) if self._storage is not None else None

    def update_interface_settings(self, new_interface_settings):
        interface_settings_flag = False
        interface_data_flag = False
        try:
            if new_interface_settings is not None:
                interface_settings_flag = interface_settings_are_changed(self._storage[STORAGE_INTERFACE_SETTINGS],
                                                                         new_interface_settings)
                if interface_settings_flag:
                    prev_storage = self._storage.copy()
                    self.logger.info("PREV INTERFACE SETTINGS:\n{0}".format(prev_storage[STORAGE_INTERFACE_SETTINGS]))
                    self.logger.info("NEW INTERFACE SETTINGS:\n{0}".format(new_interface_settings))

                    self._storage = self._storage_init(new_interface_settings)
                    prev_storage.pop(STORAGE_PREV, None)
                    self._storage[STORAGE_PREV] = prev_storage

                    self._set_storage_file()

                    interface_data_flag = self._interface_data_is_changed(
                        self._storage[STORAGE_INTERFACE_DATA],
                        self._storage[STORAGE_PREV][STORAGE_INTERFACE_DATA]
                    )
        except:
            self.logger.error(traceback.format_exc())
        return interface_settings_flag, interface_data_flag

    # def set_interface_settings(self, new_interface_settings):
    #     if new_interface_settings is None:
    #         new_interface_settings = S.DEFAULT_INTERFACE_SETTINGS
    #
    #     prev_storage = self._storage.copy()
    #     self.logger.info("PREV INTERFACE SETTINGS:\n{0}".format(prev_storage[STORAGE_INTERFACE_SETTINGS]))
    #     self.logger.info("NEW INTERFACE SETTINGS:\n{0}".format(new_interface_settings))
    #
    #     self._storage = self._storage_init(new_interface_settings)
    #     prev_storage.pop(STORAGE_PREV, None)
    #     self._storage[STORAGE_PREV] = prev_storage
    #
    #     self._set_storage_file()
    #     return self._interface_data_is_changed(self._storage[STORAGE_INTERFACE_DATA],
    #                                            self._storage[STORAGE_PREV][STORAGE_INTERFACE_DATA])

    def interface_data(self):
        return self._storage[STORAGE_INTERFACE_DATA] if self._storage is not None else None

    def _get_interface_data(self, interface_settings):
        result_interface_settings = {i: S.INTERFACE_SETTINGS_TEMPLATE[i][0] for i in S.BASE_INTERFACES}
        # i_name -> i_data(txt)
        interface_priorities = {}

        for interface_name, client_interface_settings in interface_settings.items():
            init_interface_settings = S.INTERFACE_SETTINGS_TEMPLATE.get(interface_name)
            if init_interface_settings is None:
                continue
            is_active = client_interface_settings.get('is_active', True)

            client_interface_data = client_interface_settings.get('data')
            if (not is_active) or ((client_interface_data is not None) and (not client_interface_data)):
                # if not active or empty string interface_data
                result_interface_settings.pop(interface_name, None)
                continue

            if client_interface_data is not None:
                interface_data = client_interface_data
            else:
                interface_data = init_interface_settings[0]
                if interface_name == S.WLAN_INTERFACE_NAME:
                    ssid = client_interface_settings.get('ssid')
                    if (ssid is None) or (not ssid):
                        self.logger.error("There is no \"ssid\" WLAN in the settings json-file")
                        continue
                    key = client_interface_settings.get('key')
                    if (key is not None) and key:
                        interface_data = interface_data.format(ssid, key)
                    else:
                        interface_data = S.FREE_WLAN_INTERFACE_DATA.format(ssid)
                        self.logger.warning("There is no \"key\" for WLAN")

            if interface_data is not None:
                result_interface_settings[interface_name] = interface_data
                interface_priorities[interface_name] = client_interface_settings.get('priority',
                                                                                     S.DEFAULT_INTERFACE_PRIORITY)

        # self._ppp_is_active = PPP_INTERFACE_NAME in interface_settings
        # for i in list(interface_settings):
        #     if i == PPP_INTERFACE_NAME:
        #         self._ppp_is_active = True
        #         if self._net.is_4g_modem():
        #             interface_settings[USB0_INTERFACE_NAME] = INTERFACE_SETTINGS_TEMPLATE[USB0_INTERFACE_NAME][0]
        #             interface_priorities[USB0_INTERFACE_NAME] = interface_priorities.get(i, DEFAULT_INTERFACE_PRIORITY)
        #             interface_settings.pop(i, None)
        #             interface_priorities.pop(i, None)

        prioritized_interfaces = [(i, d, interface_priorities.get(i, S.DEFAULT_INTERFACE_PRIORITY),
                                   S.INTERFACE_SETTINGS_TEMPLATE[i][1]) for i, d in result_interface_settings.items()]
        prioritized_interfaces.sort(key=lambda x: (x[2], x[3]))  # sort by pair (PRIORITY, ORDER)

        result_interfaces_data = []
        metric = 0
        for (i, d, p, o) in prioritized_interfaces:
            if i == S.LO_INTERFACE_NAME:
                result_interfaces_data.insert(0, [i, d, 0])
                continue
            result_interfaces_data.append([i, d, metric])
            metric += 10
        return result_interfaces_data

    def _interface_data_is_changed(self, interface_settings1, interface_settings2):
        try:
            interface_set1 = set()
            interface_set2 = set()
            for (i_settings, i_set) in [(interface_settings1, interface_set1), (interface_settings2, interface_set2)]:
                for (i, d, m) in i_settings:
                    i_set.add((i, d))
            if interface_set1 == interface_set2:
                return False
        except:
            self.logger.error(traceback.format_exc())
        return True
