import json
import traceback
import threading

import paho.mqtt.client as mqtt

from .sys_utils import IQLog, except_log
from . import settings as S


def interface_settings_are_changed(settings1, settings2):
    if (settings1 is not None) and (settings2 is not None):
        return settings1 != settings2
    if (settings1 is None) and (settings2 is None):
        return False
    return True


class MQTTClient(IQLog):
    def __init__(self, logger=None):
        super().__init__(logger)

        self._connected_flag = False
        self._settings_mutex = threading.Lock()
        self._settings = None
        self._new_settings_flag = False

        self._client = mqtt.Client()
        self._message_callbacks = [
            (S.SETTINGS_MQTT_TOPIC, self._on_settings)
        ]

        self._client.on_connect = self._on_connect
        self._client.on_disconnect = self._on_disconnect

    @except_log
    def _on_settings(self, client, userdata, msg):
        settings = self._process_json_payload(msg)
        # if settings is None:
        #     settings = {}
        self.logger.info("OLD SETTINGS: {0}".format(self._settings))
        self.logger.info("NEW SETTINGS: {0}".format(settings))
        with self._settings_mutex:
            if interface_settings_are_changed(self._settings, settings):
                self._settings = settings
                self._new_settings_flag = True
                self.logger.info("NEW SETTINGS - OK!".format(settings))

    def new_settings(self):
        return self._new_settings_flag

    def get_settings(self):
        with self._settings_mutex:
            self._new_settings_flag = False
            if self._settings is None:
                return None
            return self._settings.copy()

    @except_log
    def start(self):
        self.logger.info("Connect to mqtt")
        # while not self._connected_flag:
        self._client.connect('localhost')
            # time.sleep(5)
        self.logger.info("Connect to mqtt ... OK")

        self._client.loop_forever()  # loop_start()
        self.logger.info("FINISH")

    @except_log
    def _on_connect(self, client, userdata, flags, rc):
        txt = "Connection returned result: {0}".format(mqtt.connack_string(rc))
        if rc != mqtt.CONNACK_ACCEPTED:
            self._connected_flag = False  # Just in case
            self.logger.error(txt)
        else:
            self._connected_flag = True
            self.logger.info(txt)
        if (self._message_callbacks is not None) and self._message_callbacks:
            for topic, callback in self._message_callbacks:
                self._client.message_callback_add(topic, callback)
                self._client.subscribe(topic, 0)

    @except_log
    def _on_disconnect(self, client, userdata, rc):
        self._connected_flag = False
        txt = "Disconnect returned result: {0}".format(mqtt.error_string(rc))
        if rc != 0:
            self.logger.error("{0} [{1}]".format(txt, "Unexpected disconnection"))
            # self.client.reconnect()  # loop_start()/loop_forever() automatically handles reconnecting
        else:
            self.logger.warning(txt)  # disconnect by disconnect() method of client

    def _process_json_payload(self, msg):
        try:
            payload = json.loads(msg.payload.decode("utf-8"))
        except:
            payload = None
            self.logger.error(traceback.format_exc())
        return payload

    def _publish(self, topic, payload=None, qos=2, retain=True):
        try:
            # self.logger.info("start publish")
            publish_info = self._client.publish(topic, payload, qos, retain)
            # self.logger.info("publish ... ok")
            result = True if publish_info.rc == mqtt.MQTT_ERR_SUCCESS else False
        except:
            result = False
            self.logger.error(traceback.format_exc())
        if not result:
            err_txt = "Cannot send message with topic [{0}]: \"{1}\"".format(topic, payload)
            self.logger.error(err_txt)
        return result

    def _publish_as_json(self, topic, payload=None, qos=2, retain=True):
        try:
            # self.logger.info("get json_payload")
            json_payload = json.dumps(payload, separators=(',', ':'), sort_keys=True)
            # self.logger.info("json_payload ... ok")
        except:
            self.logger.error(traceback.format_exc())
            return False
        return self._publish(topic, json_payload, qos, retain)

    def publish_result(self, result):
        return self._publish_as_json(S.RESULT_MQTT_TOPIC, result)
