import os
import shutil
import datetime
import time
import traceback

from .sys_utils import ShellCommand, except_log
from .gsm_modem import GSMModem

from . import settings as S


PUBLIC_DNS = [
    "1.1.1.1", "1.0.0.1",
    "84.200.69.80", "84.200.70.40",
    "8.8.8.8", "8.8.4.4",
    "208.67.222.222", "208.67.220.22"
]


def check_dns():
    resolvconf_dir = "/etc/resolvconf/resolv.conf.d"
    resolv_conf_filepath = os.path.join(resolvconf_dir, "tail")
    if not os.path.isdir(resolvconf_dir):
        return False

    new_dns = set(PUBLIC_DNS)
    i = 0
    try:
        with open(resolv_conf_filepath) as file:
            for line in file:
                i += 1
                line = line.strip()
                if line and (not line.startswith("#")):
                    for dns in PUBLIC_DNS:
                        if dns in line:
                            try:
                                new_dns.remove(dns)
                            except KeyError:
                                pass
    except FileNotFoundError:
        pass

    if new_dns:
        nameserver_t = "nameserver {0}\n"
        lines = [nameserver_t.format(dns) for dns in new_dns]
        lines.insert(0, "# [IQnergy: {0}]\n".format(datetime.datetime.now()))
        if i > 0:
            lines.insert(0, "\n")
        with open(resolv_conf_filepath, 'a') as file:
            file.writelines(lines)
    return True


class NetUtils(ShellCommand):
    def __init__(self, storage, logger=None):
        super().__init__(logger)
        self._modem = GSMModem(logger=self.logger)
        self._storage = storage
        # self.update_modem_interface_flag = False

    def print_routes(self):
        self.command_exec("route -n")

    @except_log
    def check_dns(self):
        if not check_dns():
            self.logger.error("There is problem with [resolvconf]")

    def controller_reboot(self):
        self.logger.warning("!!!    CONTROLLER REBOOT    !!!")
        self.command_exec('reboot')

    def modem_settings(self):
        return self._modem.settings()

    def get_current_gateways(self):
        result = {}
        route_filepath = "/proc/net/route"
        try:
            with open(route_filepath) as f:
                headers = {}
                header_flag = True
                for line in f:
                    elems = line.split()
                    if header_flag:
                        i = 0
                        for elem in elems:
                            headers[elem.lower()] = i
                            i += 1
                        header_flag = False

                        for key_elem in ['iface', 'destination', 'flags', 'metric']:
                            if key_elem not in headers:
                                break
                        continue

                    if len(elems) < len(headers):
                        continue

                    # if fields[1] != '00000000' or not int(fields[3], 16) & 2:
                    if (elems[headers['destination']] == '00000000') or (int(elems[headers['flags']], 16) & 2):
                        iface = elems[headers['iface']]
                        metric = int(elems[headers['metric']])
                        metrics = result.get(iface)
                        if metrics is None:
                            metrics = []
                            result[iface] = metrics
                        metrics.append(metric)
        except Exception:
            self.logger.error(traceback.format_exc())
        return result

    @except_log
    def delete_default_gateways(self, interface=None):
        self._delete_default_gateways_template(interface, lambda x, y: False if (x is not None) and (x != y) else True)

    # def delete_other_default_gateways(self, interface):
    #     self._delete_default_gateways_template(interface, lambda x, y: True if x != y else False)

    def _delete_default_gateways_template(self, input_interface, interface_filter=None):
        interfaces = {}

        gateway_interfaces = self.get_current_gateways()
        for interface_name, interface_metrics in gateway_interfaces.items():
            if (interface_filter is None) or \
                                ((interface_filter is not None) and
                                 (interface_filter(input_interface, interface_name))):
                interfaces[interface_name] = len(interface_metrics)

        for next_interface, num_gateway_routes in interfaces.items():
            for i in range(num_gateway_routes + 1):
                # print("{0}/{1}. {3} [{4}]: {2}".format(i+1, num_routes+1, next_interface, interface_filter, prefix))
                if self._del_interface_gateway(next_interface) != 0:
                    # print("BREAK")
                    break

    def _del_interface_gateway(self, interface_name):
        res, out, err = self.command_exec("route del default dev {0}".format(interface_name), timeout=60)
        # await asyncio.sleep(DELAY)
        return res

    def _ping_interface(self, hostname, interface=None, count=1, timeout_sec=5):
        cmd = "timeout {2} ping {0} -c {1} -w {2}".format(hostname, count, timeout_sec)
        if interface is not None:
            cmd += " -I {0}".format(interface)
        cmd += " > /dev/null"

        res, out, err = self.command_exec(cmd)
        if res != 0:
            return False
        return True

    def _check_internet(self, interface=None):
        test_hosts = ["www.ya.ru", "www.google.com", "wikipedia.org",
                      "online.iqnergy.com", "my.iqnergy.com"] + PUBLIC_DNS
        n = 0
        for hostname in test_hosts:
            result_flag = self._ping_interface(hostname, interface)
            if result_flag:
                n += 1
            if n > 0:
                break
            time.sleep(1)
        if n > 0:
            return True
        return False

    def internet_via_eth(self):
        checked_interfaces = []
        for interface in [S.ETH0_INTERFACE_NAME, S.ETH1_INTERFACE_NAME]:
            if skip_interface(interface):
                self.logger.warning(" - Physical link is down! Skip [{0}] interface".format(interface))
                continue
            if self._check_internet(interface):
                checked_interfaces.append(interface)
        return checked_interfaces

    def ifdown_not_actual_interfaces(self, previous_interfaces, actual_interfaces):
        for interface in previous_interfaces:
            if interface not in actual_interfaces:
                self.logger.warning("IFDOWN \"{0}\" interface".format(interface))
                self._ifdown(interface)

    def check_modem(self, target_interfaces):
        metric = target_interfaces.get(S.PPP_INTERFACE_NAME)
        if metric is None:
            # self._modem.off()  # JUST IN CASE OF WB BUGS
            self.logger.info("RETURN check_modem = True (MODEM OFF)")
            return True

        if not self._modem.check_status():
            self.logger.info("self._modem.check_status() = False")
            # time.sleep(S.DELAY_DEFAULT)
            # self.logger.info("RETRY self._modem.check_status()")
            # if not self._modem.check_status():
            #     self.logger.info("RETURN check_modem = False")
            #     return False
            return False
        self.logger.info("RETURN check_modem = True")
        return True

    def update_sim_data(self):
        return self._modem.update_sim_slot_data()

    def _change_modem_sim_slot(self):
        if self._modem.change_sim_slot():
            self.logger.info("Modem SIM slot is changed")
        else:
            self.logger.error("Modem SIM slot is NOT changed")

    def check_interface(self, interface_name, current_metrics, target_metric):
        if interface_name == S.PPP_INTERFACE_NAME:
            if self._modem.skip():
                # self._up_interface(interface_name, target_metric)
                self.logger.warning("SKIP \"{0}\" interface by modem".format(S.PPP_INTERFACE_NAME))
                return False
            if self._modem.sim_not_inserted():
                self.logger.warning("SKIP \"{0}\" interface by modem SIM not inserted".format(S.PPP_INTERFACE_NAME))
                self._change_modem_sim_slot()
                return False

        # recover_interface_flag = True
        if (len(current_metrics) == 1) and (current_metrics[0] == target_metric):
            if self._check_internet(interface_name):
                # recover_interface_flag = False
                return True
            else:
                # delete_default_gateways(interface_name)
                self.logger.warning("Interface \"{0}\": NO INTERNET".format(interface_name))
        # if recover_interface_flag:
        if self._recover_interface(interface_name, target_metric):
            self.logger.info(" - RECOVER OK! Interface [{0}]".format(interface_name))
            return True
        else:
            self.logger.warning(" - RECOVER FAILED, NO INTERNET! Interface [{0}]".format(interface_name))
        return False

    # def modem_restart(self):
    #     return self._modem.restart()
    #     # self._ifup(S.PPP_INTERFACE_NAME)
    #     # time.sleep(S.DELAY_GSM)

    # def modem_on(self):
    #     return self._modem.on_not_blocking()
    #     # return self._modem.on()

    # def modem_off(self):
    #     return self._modem.off()

    def _up_interface(self, interface, metric):
        self._ifup(interface)
        time.sleep(2*S.DELAY_DEFAULT)
        self._add_gateway(interface, metric)

    def _recover_interface(self, interface, metric=0):
        self.delete_default_gateways(interface)  # only if NO INTERNET at interface

        if skip_interface(interface):
            self.logger.warning(" - Physical link is down! Skip [{0}] interface".format(interface))
            return False

        self.logger.info(" - Stage 1 [{0}, {1}]".format(interface, metric))
        # self._add_gateway(interface, metric)
        if self._add_gateway(interface, metric):
            time.sleep(S.DELAY_DEFAULT)
            if self._check_internet(interface):
                return True

        self.logger.info(" - Stage 2 [{0}, {1}]".format(interface, metric))

        self._ifdown(interface)
        self._up_interface(interface, metric)
        # self.ifup(interface)
        # self._add_gateway(interface, metric)

        # if interface == PPP_INTERFACE_NAME:
        #     await self.add_ppp_gateway(interface, metric=metric)
        # else:
        #     await self.dhclient(interface, metric=metric)

        if self._check_internet(interface):
            return True

        # if ((interface == S.PPP_INTERFACE_NAME) or (interface == USB0_INTERFACE_NAME)) and (self._modem is not None):
        if interface == S.PPP_INTERFACE_NAME:
            # self.logger.info(" - Stage 3 (GSM reset)")
            # self._ifdown(interface)
            # self._modem.off()
            # # self._modem.on()
            # self._up_interface(interface, metric)
            # # self.ifup(interface)
            # # # time.sleep(S.DEFAULT_DELAY)
            # # self._add_gateway(interface, metric)
            #
            # time.sleep(S.DELAY_NO_INTERNET)
            #
            # if self._check_internet(interface):
            #     return True

            self.logger.info(" - Stage 4 (change sim slot)")
            # self._ifdown(interface)

            self._change_modem_sim_slot()

            # new_provider_flag = self._modem.change_sim_slot()
            # if new_provider_flag:
            #     self._update_modem_interface()
            #
            # self._up_interface(interface, metric)
            # # self.ifup(interface)
            # # # time.sleep(S.DEFAULT_DELAY)
            # # self._add_gateway(interface, metric)
            # time.sleep(S.DELAY_NO_INTERNET)
            # if self._check_internet(interface):
            #     return True

            # self.update_modem_interface_flag = True
            # self._modem.restart()

            # self.ifup(interface)
            # self._add_gateway(interface, metric)
            #
            # if self.check_internet(interface):
            #     return True

        self.delete_default_gateways(interface)  # only if NO INTERNET at interface
        return False

    # def is_modem_interface_required(self):
    #     if self.update_modem_interface_flag:
    #         result = True
    #         self.update_modem_interface_flag = False
    #     else:
    #         result = False
    #     return result

    # def _update_modem_interface(self):
    #     self.logger.info("Update modem interface file")
    #     shutil.copy(S.LINUX_INTERFACES_FILE_PATH, S.LINUX_INTERFACES_BACKUP_FILE_PATH)
    #
    #     # interfaces_text = "\n\n".join([d for (i, d, m) in interface_data]) + "\n"
    #     interfaces_text = ""
    #     for (i, d, m) in self._storage.interface_data():
    #         if i == S.PPP_INTERFACE_NAME:
    #             # self.logger.info("CURRENT PROVIDER = {0}".format(self._modem.current_sim_provider()))
    #             # self.logger.info("PPP INTERFACE = {0}".format(d.format(self._modem.current_sim_provider())))
    #             interfaces_text += "\n{0}\n".format(d.format(self._modem.current_sim_provider()))
    #         else:
    #             interfaces_text += "\n{0}\n".format(d)
    #
    #     with open(S.LINUX_INTERFACES_FILE_PATH, 'w') as f:
    #         f.write(interfaces_text)
    #
    #     self.logger.info("Update modem interface file ... OK")
    #
    #     # # ifdown -i /etc/network/interfaces.bak <your_interface>; ifup <your_interface>
    #     # self.command_exec("ifdown -i {0} -a".format(S.LINUX_INTERFACES_BACKUP_FILE_PATH))
    #     # self.command_exec("ifup -a")
    #     #
    #     # # (sleep 180; ifdown <your_interface>; ifup -i /etc/network/interfaces.bak <your_interface>)&
    #     # time.sleep(S.DEFAULT_DELAY)

    def _add_gateway(self, interface, metric):
        if interface == S.PPP_INTERFACE_NAME:
            ret = self._add_ppp_gateway(interface, metric=metric)
        else:
            ret = self._dhclient(interface, metric=metric)
        if (ret is None) or (ret != 0):
            return False
        time.sleep(S.DELAY_COMMAND_RESULT)
        return True

    def _add_ppp_gateway(self, interface_name, metric=100):
        cmd = "route add default dev {0} metric {1}".format(interface_name, metric)
        ret, stdout, stderr = self.command_exec(cmd, timeout=30)
        return ret

    def _dhclient(self, interface, metric=100):
        # cmd = "dhclient -r && dhclient -1 {0}".format(interface)
        cmd = "/sbin/dhclient -1 -e IF_METRIC={1} {0}".format(interface, metric)
        # # cmd = "dhclient -1 {0}".format(interface)
        # result = command_exec(cmd)
        # time.sleep(1)
        # return result
        ret, stdout, stderr = self.command_exec(cmd, timeout=30)
        return ret

    def _ifup(self, interface):
        # cmd = "ifup --force {0}".format(interface)
        cmd = "ifup {0}".format(interface)
        result = self.command_exec(cmd, timeout=60)
        time.sleep(S.DELAY_COMMAND_RESULT)
        return result

    def _ifdown(self, interface):
        cmd = "ifdown --force {0}".format(interface)
        return self.command_exec(cmd, timeout=60)

    def update_linux_interfaces(self):
        shutil.copy(S.LINUX_INTERFACES_FILE_PATH, S.LINUX_INTERFACES_BACKUP_FILE_PATH)

        # interfaces_text = "\n\n".join([d for (i, d, m) in interface_data]) + "\n"
        # ppp_inserted_flag = False
        interfaces_text = []
        for (i, d, m) in self._storage.interface_data():
            if i == S.PPP_INTERFACE_NAME:
                # self.logger.info("CURRENT PROVIDER = {0}".format(self._modem.current_sim_provider()))
                # self.logger.info("PPP INTERFACE = {0}".format(d.format(self._modem.current_sim_provider())))
                interfaces_text.append(d.format(self._modem.current_ppp_provider()).strip())
                # ppp_inserted_flag = True
            else:
                interfaces_text.append(d.strip())
                # if not ppp_inserted_flag:
                #     interfaces_text.append(d.strip())
                # else:
                #     interfaces_text.insert(len(interfaces_text)-1, d.format(self._modem.current_ppp_provider()).strip())

        with open(S.LINUX_INTERFACES_FILE_PATH, 'w') as f:
            f.write("\n\n\n".join(interfaces_text)+'\n\n')

        # ifdown -i /etc/network/interfaces.bak <your_interface>; ifup <your_interface>
        # self.command_exec("ifdown -i {0} -a".format(S.LINUX_INTERFACES_BACKUP_FILE_PATH), timeout=60)
        # self.command_exec("ifup -a", timeout=60)

        # (sleep 180; ifdown <your_interface>; ifup -i /etc/network/interfaces.bak <your_interface>)&
        time.sleep(S.DELAY_COMMAND_RESULT)

    # async def restore_interface_settings(self, interface_settings):
    #     if not os.path.isfile(LINUX_INTERFACES_BACKUP_FILE_PATH):
    #         return False
    #
    #     await self.command_exec("ifdown -a")
    #     await self.command_exec("ifup -i {0} -a".format(LINUX_INTERFACES_BACKUP_FILE_PATH))
    #     return True

    # @staticmethod
    # def _parse_wlan_quality_param(param_str):
    #     # param_str example: 'Signal level=100/100  '
    #     tmp = param_str.split('=')
    #     if len(tmp) > 1:
    #         return tmp[-1].strip()

    def _wlan_quality(self):
        cmd = "iwconfig wlan0 | grep -i quality"
        ret, out, err = self.command_exec(cmd, timeout=15)
        if (ret is None) or (ret != 0):
            return None

        # stdout example: '          Link Quality=98/100  Signal level=100/100  Noise level=0/100'
        # out = out.strip().lower().split('  ')
        result = {}
        for part in out.strip().lower().split('  '):
            r = part.split('=')
            if len(r) != 2:
                continue
            result[r[0]] = r[1]

        # params = ["link quality", "signal level", "noise level"]
        # tmp = []
        # for param in params:
        #     i = out.find(param)
        #     if i < 0:
        #         continue
        #     tmp.append((i, param))
        # if not tmp:
        #     return
        #
        # result = {}
        # tmp.sort(key=lambda x: x[0])
        # tmp_len = len(tmp)
        # if tmp_len > 1:
        #     for i in range(tmp_len-1):
        #         result[tmp[i][1]] = self._parse_wlan_quality_param(out[tmp[i][0]:tmp[i+1][0]])
        # result[tmp[-1][1]] = self._parse_wlan_quality_param(out[tmp[-1][0]:])
        return result

    def interface_quality(self, checked_interfaces):
        result = {}
        for interface in checked_interfaces:
            if interface == S.PPP_INTERFACE_NAME:  # and (not self._modem.skip()): - NOT NEEDED checked (!) interfaces
                # or (interface == S.USB0_INTERFACE_NAME):
                quality = self._modem.signal_quality()
            elif interface == S.WLAN_INTERFACE_NAME:
                quality = self._wlan_quality()
            elif interface in [S.ETH0_INTERFACE_NAME, S.ETH1_INTERFACE_NAME]:
                quality = True
            else:
                continue
            if (quality is not None) and quality:
                result[interface] = quality
        return result


def skip_interface(interface):
    try:
        if interface in [S.ETH0_INTERFACE_NAME, S.ETH1_INTERFACE_NAME]:  # , WLAN_INTERFACE_NAME]:
            # File indicates the current physical link state of the interface.
            filepath = "/sys/class/net/{0}/carrier".format(interface)
            if os.path.isfile(filepath):
                with open(filepath, 'r') as file:
                    state_txt = file.read()
                    try:
                        state = int(state_txt.strip())
                        if state == 0:
                            return True
                    except ValueError:
                        pass
    except Exception:
        pass
    return False


###################################################################################################


# def rx_tx_bytes(interfaces=None):
#     with open("/proc/net/dev", 'r') as file:
#         lines = file.readlines()
#     if not lines or len(lines) < 2:
#         return {}
#     columnLine = lines[1]
#     _, receiveCols, transmitCols = columnLine.split("|")
#     cols = []
#     for pref, tmp_cols in [('recv_', receiveCols.split()), ('trans_', transmitCols.split())]:
#         cols += [pref+field for field in tmp_cols]
#     result = {}
#     for line in lines[2:]:
#         if line.find(":") < 0:
#             continue
#         interface, data = line.split(":")
#         interface = interface.strip()
#         if (interfaces is not None) and (interface not in interfaces):
#             continue
#         interface_data = dict(zip(cols, [int(d) for d in data.split()]))
#         result[interface] = interface_data
#     return result
#
#
# class BytesCounter:
#     def __init__(self):
#         self._stat = {}  # interface: (timestamp, recv, trans)
#     def count(self, interfaces=None):
#         result = {}
#         ts = datetime.datetime.now().timestamp()
#         tmp = rx_tx_bytes(interfaces)
#         for interface, interface_data in tmp.items():
#             current_stat = ts, interface_data.get('recv_bytes'), interface_data.get('trans_bytes')
#             prev_stat = self._stat.get(interface)
#             if prev_stat is None:
#                 self._stat[interface] = current_stat
#                 continue
#             d = ts - prev_stat[0]
#             if d <= 0:
#                 continue
#             if d > 2*60:
#                 self._stat[interface] = current_stat
#                 continue
#             self._stat[interface] = current_stat
#             r = None
#             if (current_stat[1] is not None) and (prev_stat[1] is not None):
#                 r = (current_stat[1]-prev_stat[1]) / d
#                 if r < 0:
#                     r = None
#             t = None
#             if (current_stat[2] is not None) and (prev_stat[2] is not None):
#                 t = (current_stat[2]-prev_stat[2]) / d
#                 if t < 0:
#                     t = None
#             result[interface] = ts, r, t
#         return result
