import os
import sys
import importlib.util
import logging
import logging.handlers
from functools import wraps
import subprocess
import traceback


PATH_HEADER = "PATH=$(su - root -c \"echo \\$PATH\")\n"

LOG_FORMAT = " %(asctime)s | %(levelname)8s | [%(module)s]:[%(funcName)s] %(message)s"
LOG_TIME_FORMAT = '%d.%m.%Y %H:%M:%S'

LOG_BACKUP_COUNT = 24


def get_logger(filepath, name=None):
    try:
        new_log = logging.getLogger(name)
        new_log.setLevel(logging.INFO)

        for handler in reversed(new_log.handlers):
            handler.close()
            new_log.removeHandler(handler)

        handler = logging.handlers.TimedRotatingFileHandler(filepath,
                                                            when='h', interval=1, backupCount=LOG_BACKUP_COUNT)
        formatter = logging.Formatter(fmt=LOG_FORMAT, datefmt=LOG_TIME_FORMAT)
        handler.setFormatter(formatter)

        new_log.addHandler(handler)
        new_log.addHandler(logging.StreamHandler(sys.stdout))

        new_log.propagate = False
    except:
        print(traceback.format_exc())
        new_log = get_null_logger()
    return new_log


def get_null_logger():
    new_logger = logging.getLogger('null')
    new_logger.setLevel(logging.INFO)

    for handler in reversed(new_logger.handlers):
        handler.close()
        new_logger.removeHandler(handler)

    new_logger.addHandler(logging.NullHandler())
    new_logger.propagate = False
    return new_logger


def import_module(dir_path, file_name):
    """
    Функция, возвращающая python-модуль по его пути
    """
    file_path = os.path.join(dir_path, file_name)
    module_name, ext = os.path.splitext(file_name)

    spec = importlib.util.spec_from_file_location(module_name, file_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def except_log(function):
    @wraps(function)
    def wrapper(*args, **kwargs):
        try:
            return function(*args, **kwargs)
        except:
            args[0].logger.critical(traceback.format_exc())
    return wrapper


class IQLog:
    def __init__(self, logger=None):
        self.logger = logger if logger is not None else get_null_logger()


class ShellCommand(IQLog):
    def __init__(self, logger=None):
        super().__init__(logger)

    def command_exec(self, cmd, timeout=None):
        ret = None
        stdout = ''
        stderr = ''
        try:
            cp = subprocess.run(cmd,
                                shell=True, executable='/bin/bash',
                                stdout=subprocess.PIPE, stderr=subprocess.PIPE,
                                timeout=timeout)
            ret = cp.returncode
            stdout = self._get_text(cp.stdout)
            stderr = self._get_text(cp.stderr)

            log_text = "COMMAND \"{0}\" exited with {1}".format(cp.args, cp.returncode)
            if stdout:
                log_text += "\n[stdout]\n\"{0}\"".format(stdout)
            if stderr:
                log_text += "\n[stderr]\n\"{0}\"".format(stderr)
            self.logger.info(log_text)
        except subprocess.TimeoutExpired:
            self.logger.warning("COMMAND TIMEOUT: \"{0}\"".format(cmd))
        except:
            self.logger.error(traceback.format_exc())
        return ret, stdout, stderr

    def command_exec_background(self, cmd):
        try:
            p = subprocess.Popen(cmd, shell=True, executable='/bin/bash',
                                 stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        except:
            self.logger.error(traceback.format_exc())
            return None
        return p

    def command_strip_output(self, cmd):
        ret, out, err = self.command_exec(cmd)
        return out.strip()

    def _get_text(self, bytes_sequence):
        try:
            result = bytes_sequence.decode('utf-8')
        except:
            self.logger.error(traceback.format_exc())
            result = ''
        return result
